from typing import Optional
import logging
import time
from torch import Tensor
from pytorch_lightning.callbacks.base import Callback

import graphsignal
from graphsignal.proto import profiles_pb2
from graphsignal.profilers.pytorch import PyTorchProfiler
from graphsignal.profiling_step import ProfilingStep

logger = logging.getLogger('graphsignal')

PHASE_TRAINING = 'training'
PHASE_VALIDATION = 'validation'
PHASE_TEST = 'test'
PHASE_PREDICTION = 'prediction'


class GraphsignalCallback(Callback):
    def __init__(self, batch_size: Optional[int] = None):
        super().__init__()
        self._profiler = PyTorchProfiler()
        self._step = None
        self._batch_size = batch_size
        self._node_rank = -1
        self._local_rank = -1
        self._global_rank = -1

    def on_train_start(self, trainer, pl_module):
        self._configure_profiler(trainer)

    def on_train_end(self, trainer, pl_module):
        self._stop_profiler(trainer)

    def on_validation_start(self, trainer, pl_module):
        self._configure_profiler(trainer)

    def on_validation_end(self, trainer, pl_module):
        self._stop_profiler(trainer)

    def on_test_start(self, trainer, pl_module):
        self._configure_profiler(trainer)

    def on_test_end(self, trainer, pl_module):
        self._stop_profiler(trainer)

    def on_predict_start(self, trainer, pl_module):
        self._configure_profiler(trainer)

    def on_predict_end(self, trainer, pl_module):
        self._stop_profiler(trainer)

    def on_train_batch_start(self, trainer, pl_module, batch, batch_idx):
        self._stop_profiler(trainer)
        self._start_profiler(PHASE_TRAINING, trainer)

    def on_train_batch_end(self, trainer, pl_module, outputs, batch, batch_idx):
        self._log_metrics(trainer.logged_metrics)

    def on_validation_batch_start(self, trainer, pl_module, batch, batch_idx, dataloader_idx):
        self._stop_profiler(trainer)
        self._start_profiler(PHASE_VALIDATION, trainer)

    def on_validation_batch_end(self, trainer, pl_module, outputs, batch, batch_idx, dataloader_idx):
        self._log_metrics(trainer.logged_metrics)

    def on_test_batch_start(self, trainer, pl_module, batch, batch_idx, dataloader_idx):
        self._stop_profiler(trainer)
        self._start_profiler(PHASE_TEST, trainer)

    def on_test_batch_end(self, trainer, pl_module, outputs, batch, batch_idx, dataloader_idx):
        self._log_metrics(trainer.logged_metrics)

    def on_predict_batch_start(self, trainer, pl_module, batch, batch_idx, dataloader_idx):
        self._stop_profiler(trainer)
        self._start_profiler(PHASE_PREDICTION, trainer)

    def on_predict_batch_end(self, trainer, pl_module, outputs, batch, batch_idx, dataloader_idx):
        self._log_metrics(trainer.logged_metrics)

    def _configure_profiler(self, trainer):
        if self._check_param(trainer, 'node_rank') and trainer.node_rank >= 0:
            self._node_rank = trainer.node_rank
        if self._check_param(trainer, 'local_rank') and trainer.local_rank >= 0:
             self._local_rank = trainer.local_rank
        if self._check_param(trainer, 'global_rank') and trainer.global_rank >= 0:
             self._global_rank = trainer.global_rank

        if self._batch_size:
            graphsignal.log_parameter('batch_size', self._batch_size)

        self._log_basic_param(trainer, 'auto_scale_batch_size')
        self._log_basic_param(trainer, 'accumulate_grad_batches')
        self._log_basic_param(trainer, 'world_size')
        self._log_basic_param(trainer, 'num_nodes')
        self._log_basic_param(trainer, 'num_devices')
        self._log_basic_param(trainer, 'precision')
        self._log_basic_param(trainer, 'data_parallel')
        self._log_basic_param(trainer, 'max_epochs')
        self._log_basic_param(trainer, 'min_epochs')
        self._log_basic_param(trainer, 'max_steps')
        self._log_basic_param(trainer, 'min_steps')

    def _start_profiler(self, phase_name, trainer):
        if not self._step:
            self._step = ProfilingStep(
                phase_name=phase_name,
                effective_batch_size=self._batch_size,
                framework_profiler=self._profiler)

    def _stop_profiler(self, trainer):
        if self._step:
            if self._step._is_scheduled:
                self._update_profile(trainer)
            self._step.stop()
            self._step = None

    def _update_profile(self, trainer):
        profile = self._step._profile

        if self._batch_size:
            profile.step_stats.batch_size = self._batch_size
            strategy = getattr(trainer, 'strategy', None)
            if strategy:
                try:
                    from pytorch_lightning.strategies import DataParallelStrategy, DDP2Strategy
                    if isinstance(strategy, (DataParallelStrategy, DDP2Strategy)):
                        num_devices = getattr(trainer, 'num_devices', 1)
                        if isinstance(num_devices, int) and num_devices > 1:
                            profile.step_stats.device_batch_size = int(self._batch_size / num_devices)
                except:
                    logger.warning('Error recording per-device batch size', exc_info=True)

        if self._check_param(trainer, 'world_size'):
            profile.step_stats.world_size = trainer.world_size

        if self._node_rank >= 0 and graphsignal._agent.node_rank == -1:
            profile.node_usage.node_rank = self._node_rank
        if self._global_rank >= 0 and graphsignal._agent.global_rank == -1:
            profile.process_usage.global_rank = self._global_rank
        if self._local_rank >= 0 and graphsignal._agent.local_rank == -1:
            profile.process_usage.local_rank = self._local_rank

    def _log_basic_param(self, trainer, param):
        value = getattr(trainer, param, None)
        if isinstance(value, (str, int, float, bool)):
            graphsignal.log_parameter(param, value)

    def _check_param(self, trainer, param):
        value = getattr(trainer, param, None)
        return isinstance(value, (str, int, float, bool))

    def _log_metrics(self, logged_metrics):
        if logged_metrics:
            for key, value in logged_metrics.items():
                if isinstance(key, str):
                    if isinstance(value, (int, float)):
                        graphsignal.log_metric(key, value)
                    elif isinstance(value, Tensor):
                        graphsignal.log_metric(key, value.item())



from graphsignal.uploader import Uploader
from graphsignal.usage.process_reader import ProcessReader
from graphsignal.usage.nvml_reader import NvmlReader

class Agent:
    worker_id: str = None
    start_ms: int = None
    api_key: str = None
    workload_name: str = None
    run_id: str = None
    global_rank: int = None
    node_rank: int = None
    local_rank: int = None
    debug_mode: bool = None
    disable_fwk_profiler: bool = False
    uploader: Uploader = None
    process_reader: ProcessReader = None
    nvml_reader: NvmlReader = None
    tags: dict = None
    params: dict = None
    metrics: dict = None



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: profiles.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0eprofiles.proto\x12\x14graphsignal.profiles\"X\n\rUploadRequest\x12\x34\n\x0bml_profiles\x18\x01 \x03(\x0b\x32\x1f.graphsignal.profiles.MLProfile\x12\x11\n\tupload_ms\x18\x02 \x01(\x04\"\x10\n\x0eUploadResponse\"\xc2\x01\n\rProfileRecord\x12\x0f\n\x07\x64\x61ta_id\x18\x01 \x01(\t\x12\x13\n\x0bworkload_id\x18\x02 \x01(\t\x12\x10\n\x08phase_id\x18\x07 \x01(\t\x12\x12\n\nprofile_id\x18\x03 \x01(\t\x12\x33\n\nml_profile\x18\x04 \x01(\x0b\x32\x1f.graphsignal.profiles.MLProfile\x12\x1a\n\x12\x64\x61ta_retention_sec\x18\x05 \x01(\x04\x12\x14\n\x0ctime_skew_ms\x18\x06 \x01(\x12\"\xf4\x05\n\tMLProfile\x12\x15\n\rworkload_name\x18\x01 \x01(\t\x12\x12\n\nphase_name\x18\x1d \x01(\t\x12\x11\n\tworker_id\x18\x1c \x01(\t\x12\x0e\n\x06run_id\x18\x02 \x01(\t\x12\x10\n\x08start_us\x18\x05 \x01(\x04\x12\x0e\n\x06\x65nd_us\x18\x06 \x01(\x04\x12\'\n\x04tags\x18  \x03(\x0b\x32\x19.graphsignal.profiles.Tag\x12+\n\x06params\x18\x16 \x03(\x0b\x32\x1b.graphsignal.profiles.Param\x12-\n\x07metrics\x18\x1e \x03(\x0b\x32\x1c.graphsignal.profiles.Metric\x12\x33\n\nstep_stats\x18\x15 \x01(\x0b\x32\x1f.graphsignal.profiles.StepStats\x12<\n\ncomm_usage\x18\x1b \x01(\x0b\x32(.graphsignal.profiles.CommunicationUsage\x12\x33\n\nnode_usage\x18\x19 \x01(\x0b\x32\x1f.graphsignal.profiles.NodeUsage\x12\x39\n\rprocess_usage\x18\x1a \x01(\x0b\x32\".graphsignal.profiles.ProcessUsage\x12\x37\n\x0c\x64\x65vice_usage\x18\x12 \x03(\x0b\x32!.graphsignal.profiles.DeviceUsage\x12/\n\x08op_stats\x18\x0b \x03(\x0b\x32\x1d.graphsignal.profiles.OpStats\x12\x37\n\x0ckernel_stats\x18\x0c \x03(\x0b\x32!.graphsignal.profiles.KernelStats\x12\x12\n\ntrace_data\x18! \x01(\x0c\x12\x1a\n\x12is_trace_available\x18\" \x01(\x08\x12<\n\x0fprofiler_errors\x18\x07 \x03(\x0b\x32#.graphsignal.profiles.ProfilerError\"5\n\x06SemVer\x12\r\n\x05major\x18\x01 \x01(\x05\x12\r\n\x05minor\x18\x02 \x01(\x05\x12\r\n\x05patch\x18\x03 \x01(\x05\"\x14\n\x03Tag\x12\r\n\x05value\x18\x01 \x01(\t\"$\n\x05Param\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t\"%\n\x06Metric\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\x01\"\xa3\x01\n\tStepStats\x12\x12\n\nstep_count\x18\x01 \x01(\x04\x12\x14\n\x0csample_count\x18\x07 \x01(\x04\x12\x12\n\nflop_count\x18\x08 \x01(\x04\x12\x15\n\rtotal_time_us\x18\x02 \x01(\x04\x12\x12\n\nbatch_size\x18\x04 \x01(\x05\x12\x19\n\x11\x64\x65vice_batch_size\x18\x05 \x01(\x05\x12\x12\n\nworld_size\x18\x06 \x01(\x05\"\xd5\x01\n\x12\x43ommunicationUsage\x12W\n\x0c\x62\x61\x63kend_type\x18\x01 \x01(\x0e\x32\x41.graphsignal.profiles.CommunicationUsage.CommunicationBackendType\"f\n\x18\x43ommunicationBackendType\x12#\n\x1f\x43OMMUNICATION_BACKEND_UNDEFINED\x10\x00\x12\x08\n\x04NCCL\x10\x01\x12\x08\n\x04GLOO\x10\x02\x12\x07\n\x03MPI\x10\x03\x12\x08\n\x04RING\x10\x04\"\xc6\x01\n\tNodeUsage\x12\x10\n\x08hostname\x18\x01 \x01(\t\x12\x12\n\nip_address\x18\x07 \x01(\t\x12\x11\n\tnode_rank\x18\x08 \x01(\x05\x12\x10\n\x08mem_used\x18\t \x01(\x04\x12\x11\n\tmem_total\x18\n \x01(\x04\x12\x10\n\x08platform\x18\x02 \x01(\t\x12\x0f\n\x07machine\x18\x03 \x01(\t\x12\x0f\n\x07os_name\x18\x04 \x01(\t\x12\x12\n\nos_version\x18\x05 \x01(\t\x12\x13\n\x0bnum_devices\x18\x06 \x01(\x05\"\xe9\x04\n\x0cProcessUsage\x12\x12\n\nprocess_id\x18\x05 \x01(\t\x12\x10\n\x08start_ms\x18\x0e \x01(\x04\x12\x12\n\nlocal_rank\x18\x0c \x01(\x05\x12\x13\n\x0bglobal_rank\x18\x0f \x01(\x05\x12\x19\n\x11\x63pu_usage_percent\x18\x01 \x01(\x01\x12\x0f\n\x07max_rss\x18\x02 \x01(\x04\x12\x13\n\x0b\x63urrent_rss\x18\x03 \x01(\x04\x12\x0f\n\x07vm_size\x18\x04 \x01(\x04\x12;\n\x07runtime\x18\x06 \x01(\x0e\x32*.graphsignal.profiles.ProcessUsage.Runtime\x12\x35\n\x0fruntime_version\x18\x07 \x01(\x0b\x32\x1c.graphsignal.profiles.SemVer\x12\x14\n\x0cruntime_impl\x18\x08 \x01(\t\x12\x44\n\x0cml_framework\x18\t \x01(\x0e\x32..graphsignal.profiles.ProcessUsage.MLFramework\x12:\n\x14ml_framework_version\x18\n \x01(\x0b\x32\x1c.graphsignal.profiles.SemVer\x12\x36\n\x10profiler_version\x18\x0b \x01(\x0b\x32\x1c.graphsignal.profiles.SemVer\",\n\x07Runtime\x12\x15\n\x11RUNTIME_UNDEFINED\x10\x00\x12\n\n\x06PYTHON\x10\x01\"F\n\x0bMLFramework\x12\x1a\n\x16ML_FRAMEWORK_UNDEFINED\x10\x00\x12\x0e\n\nTENSORFLOW\x10\x01\x12\x0b\n\x07PYTORCH\x10\x02\"\xd4\x03\n\x0b\x44\x65viceUsage\x12\x35\n\x0b\x64\x65vice_type\x18\x0b \x01(\x0e\x32 .graphsignal.profiles.DeviceType\x12\x11\n\tdevice_id\x18\x01 \x01(\t\x12\x13\n\x0b\x64\x65vice_name\x18\x02 \x01(\t\x12\x14\n\x0c\x61rchitecture\x18\x11 \x01(\t\x12\x38\n\x12\x63ompute_capability\x18\x0f \x01(\x0b\x32\x1c.graphsignal.profiles.SemVer\x12\x11\n\tmem_total\x18\x03 \x01(\x04\x12\x10\n\x08mem_used\x18\x04 \x01(\x04\x12\x10\n\x08mem_free\x18\x05 \x01(\x04\x12\x1f\n\x17gpu_utilization_percent\x18\x06 \x01(\x01\x12\x1f\n\x17mem_utilization_percent\x18\x07 \x01(\x01\x12\x1a\n\x12pcie_throughput_tx\x18\r \x01(\x01\x12\x1a\n\x12pcie_throughput_rx\x18\x0e \x01(\x01\x12\x12\n\ngpu_temp_c\x18\x08 \x01(\x01\x12\x15\n\rpower_usage_w\x18\t \x01(\x01\x12\x19\n\x11\x66\x61n_speed_percent\x18\n \x01(\x01\x12\x1f\n\x17mxu_utilization_percent\x18\x0c \x01(\x01\"\xea\x03\n\x07OpStats\x12\x35\n\x0b\x64\x65vice_type\x18\x01 \x01(\x0e\x32 .graphsignal.profiles.DeviceType\x12\x11\n\tdevice_id\x18\x02 \x01(\t\x12\x0f\n\x07op_type\x18\x03 \x01(\t\x12\x0f\n\x07op_name\x18\x04 \x01(\t\x12\r\n\x05\x63ount\x18\x05 \x01(\x04\x12\x1a\n\x12total_host_time_us\x18\x06 \x01(\x04\x12\x1c\n\x14total_device_time_us\x18\x07 \x01(\x04\x12\x19\n\x11self_host_time_us\x18\x08 \x01(\x04\x12\x1b\n\x13self_device_time_us\x18\t \x01(\x04\x12\x19\n\x11total_host_memory\x18\n \x01(\x04\x12\x1b\n\x13total_device_memory\x18\x0b \x01(\x04\x12\x18\n\x10self_host_memory\x18\x0c \x01(\x04\x12\x1a\n\x12self_device_memory\x18\r \x01(\x04\x12\x1d\n\x15self_host_memory_rate\x18\x0e \x01(\x04\x12\x1f\n\x17self_device_memory_rate\x18\x0f \x01(\x04\x12\r\n\x05\x66lops\x18\x10 \x01(\x04\x12\x15\n\rflops_per_sec\x18\x12 \x01(\x01\x12\x1e\n\x16tensorcore_utilization\x18\x11 \x01(\x01\"\xbe\x01\n\x0bKernelStats\x12\x35\n\x0b\x64\x65vice_type\x18\x01 \x01(\x0e\x32 .graphsignal.profiles.DeviceType\x12\x11\n\tdevice_id\x18\x02 \x01(\t\x12\x0f\n\x07op_name\x18\x03 \x01(\t\x12\x13\n\x0bkernel_name\x18\x04 \x01(\t\x12\r\n\x05\x63ount\x18\x05 \x01(\x04\x12\x13\n\x0b\x64uration_ns\x18\x06 \x01(\x04\x12\x1b\n\x13is_using_tensorcore\x18\x07 \x01(\x08\"5\n\rProfilerError\x12\x0f\n\x07message\x18\x01 \x01(\t\x12\x13\n\x0bstack_trace\x18\x02 \x01(\t*B\n\nDeviceType\x12\x19\n\x15\x44\x45VICE_TYPE_UNDEFINED\x10\x00\x12\x07\n\x03\x43PU\x10\x01\x12\x07\n\x03GPU\x10\x02\x12\x07\n\x03TPU\x10\x03\x62\x06proto3')

_DEVICETYPE = DESCRIPTOR.enum_types_by_name['DeviceType']
DeviceType = enum_type_wrapper.EnumTypeWrapper(_DEVICETYPE)
DEVICE_TYPE_UNDEFINED = 0
CPU = 1
GPU = 2
TPU = 3


_UPLOADREQUEST = DESCRIPTOR.message_types_by_name['UploadRequest']
_UPLOADRESPONSE = DESCRIPTOR.message_types_by_name['UploadResponse']
_PROFILERECORD = DESCRIPTOR.message_types_by_name['ProfileRecord']
_MLPROFILE = DESCRIPTOR.message_types_by_name['MLProfile']
_SEMVER = DESCRIPTOR.message_types_by_name['SemVer']
_TAG = DESCRIPTOR.message_types_by_name['Tag']
_PARAM = DESCRIPTOR.message_types_by_name['Param']
_METRIC = DESCRIPTOR.message_types_by_name['Metric']
_STEPSTATS = DESCRIPTOR.message_types_by_name['StepStats']
_COMMUNICATIONUSAGE = DESCRIPTOR.message_types_by_name['CommunicationUsage']
_NODEUSAGE = DESCRIPTOR.message_types_by_name['NodeUsage']
_PROCESSUSAGE = DESCRIPTOR.message_types_by_name['ProcessUsage']
_DEVICEUSAGE = DESCRIPTOR.message_types_by_name['DeviceUsage']
_OPSTATS = DESCRIPTOR.message_types_by_name['OpStats']
_KERNELSTATS = DESCRIPTOR.message_types_by_name['KernelStats']
_PROFILERERROR = DESCRIPTOR.message_types_by_name['ProfilerError']
_COMMUNICATIONUSAGE_COMMUNICATIONBACKENDTYPE = _COMMUNICATIONUSAGE.enum_types_by_name['CommunicationBackendType']
_PROCESSUSAGE_RUNTIME = _PROCESSUSAGE.enum_types_by_name['Runtime']
_PROCESSUSAGE_MLFRAMEWORK = _PROCESSUSAGE.enum_types_by_name['MLFramework']
UploadRequest = _reflection.GeneratedProtocolMessageType('UploadRequest', (_message.Message,), {
  'DESCRIPTOR' : _UPLOADREQUEST,
  '__module__' : 'profiles_pb2'
  # @@protoc_insertion_point(class_scope:graphsignal.profiles.UploadRequest)
  })
_sym_db.RegisterMessage(UploadRequest)

UploadResponse = _reflection.GeneratedProtocolMessageType('UploadResponse', (_message.Message,), {
  'DESCRIPTOR' : _UPLOADRESPONSE,
  '__module__' : 'profiles_pb2'
  # @@protoc_insertion_point(class_scope:graphsignal.profiles.UploadResponse)
  })
_sym_db.RegisterMessage(UploadResponse)

ProfileRecord = _reflection.GeneratedProtocolMessageType('ProfileRecord', (_message.Message,), {
  'DESCRIPTOR' : _PROFILERECORD,
  '__module__' : 'profiles_pb2'
  # @@protoc_insertion_point(class_scope:graphsignal.profiles.ProfileRecord)
  })
_sym_db.RegisterMessage(ProfileRecord)

MLProfile = _reflection.GeneratedProtocolMessageType('MLProfile', (_message.Message,), {
  'DESCRIPTOR' : _MLPROFILE,
  '__module__' : 'profiles_pb2'
  # @@protoc_insertion_point(class_scope:graphsignal.profiles.MLProfile)
  })
_sym_db.RegisterMessage(MLProfile)

SemVer = _reflection.GeneratedProtocolMessageType('SemVer', (_message.Message,), {
  'DESCRIPTOR' : _SEMVER,
  '__module__' : 'profiles_pb2'
  # @@protoc_insertion_point(class_scope:graphsignal.profiles.SemVer)
  })
_sym_db.RegisterMessage(SemVer)

Tag = _reflection.GeneratedProtocolMessageType('Tag', (_message.Message,), {
  'DESCRIPTOR' : _TAG,
  '__module__' : 'profiles_pb2'
  # @@protoc_insertion_point(class_scope:graphsignal.profiles.Tag)
  })
_sym_db.RegisterMessage(Tag)

Param = _reflection.GeneratedProtocolMessageType('Param', (_message.Message,), {
  'DESCRIPTOR' : _PARAM,
  '__module__' : 'profiles_pb2'
  # @@protoc_insertion_point(class_scope:graphsignal.profiles.Param)
  })
_sym_db.RegisterMessage(Param)

Metric = _reflection.GeneratedProtocolMessageType('Metric', (_message.Message,), {
  'DESCRIPTOR' : _METRIC,
  '__module__' : 'profiles_pb2'
  # @@protoc_insertion_point(class_scope:graphsignal.profiles.Metric)
  })
_sym_db.RegisterMessage(Metric)

StepStats = _reflection.GeneratedProtocolMessageType('StepStats', (_message.Message,), {
  'DESCRIPTOR' : _STEPSTATS,
  '__module__' : 'profiles_pb2'
  # @@protoc_insertion_point(class_scope:graphsignal.profiles.StepStats)
  })
_sym_db.RegisterMessage(StepStats)

CommunicationUsage = _reflection.GeneratedProtocolMessageType('CommunicationUsage', (_message.Message,), {
  'DESCRIPTOR' : _COMMUNICATIONUSAGE,
  '__module__' : 'profiles_pb2'
  # @@protoc_insertion_point(class_scope:graphsignal.profiles.CommunicationUsage)
  })
_sym_db.RegisterMessage(CommunicationUsage)

NodeUsage = _reflection.GeneratedProtocolMessageType('NodeUsage', (_message.Message,), {
  'DESCRIPTOR' : _NODEUSAGE,
  '__module__' : 'profiles_pb2'
  # @@protoc_insertion_point(class_scope:graphsignal.profiles.NodeUsage)
  })
_sym_db.RegisterMessage(NodeUsage)

ProcessUsage = _reflection.GeneratedProtocolMessageType('ProcessUsage', (_message.Message,), {
  'DESCRIPTOR' : _PROCESSUSAGE,
  '__module__' : 'profiles_pb2'
  # @@protoc_insertion_point(class_scope:graphsignal.profiles.ProcessUsage)
  })
_sym_db.RegisterMessage(ProcessUsage)

DeviceUsage = _reflection.GeneratedProtocolMessageType('DeviceUsage', (_message.Message,), {
  'DESCRIPTOR' : _DEVICEUSAGE,
  '__module__' : 'profiles_pb2'
  # @@protoc_insertion_point(class_scope:graphsignal.profiles.DeviceUsage)
  })
_sym_db.RegisterMessage(DeviceUsage)

OpStats = _reflection.GeneratedProtocolMessageType('OpStats', (_message.Message,), {
  'DESCRIPTOR' : _OPSTATS,
  '__module__' : 'profiles_pb2'
  # @@protoc_insertion_point(class_scope:graphsignal.profiles.OpStats)
  })
_sym_db.RegisterMessage(OpStats)

KernelStats = _reflection.GeneratedProtocolMessageType('KernelStats', (_message.Message,), {
  'DESCRIPTOR' : _KERNELSTATS,
  '__module__' : 'profiles_pb2'
  # @@protoc_insertion_point(class_scope:graphsignal.profiles.KernelStats)
  })
_sym_db.RegisterMessage(KernelStats)

ProfilerError = _reflection.GeneratedProtocolMessageType('ProfilerError', (_message.Message,), {
  'DESCRIPTOR' : _PROFILERERROR,
  '__module__' : 'profiles_pb2'
  # @@protoc_insertion_point(class_scope:graphsignal.profiles.ProfilerError)
  })
_sym_db.RegisterMessage(ProfilerError)

if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _DEVICETYPE._serialized_start=3673
  _DEVICETYPE._serialized_end=3739
  _UPLOADREQUEST._serialized_start=40
  _UPLOADREQUEST._serialized_end=128
  _UPLOADRESPONSE._serialized_start=130
  _UPLOADRESPONSE._serialized_end=146
  _PROFILERECORD._serialized_start=149
  _PROFILERECORD._serialized_end=343
  _MLPROFILE._serialized_start=346
  _MLPROFILE._serialized_end=1102
  _SEMVER._serialized_start=1104
  _SEMVER._serialized_end=1157
  _TAG._serialized_start=1159
  _TAG._serialized_end=1179
  _PARAM._serialized_start=1181
  _PARAM._serialized_end=1217
  _METRIC._serialized_start=1219
  _METRIC._serialized_end=1256
  _STEPSTATS._serialized_start=1259
  _STEPSTATS._serialized_end=1422
  _COMMUNICATIONUSAGE._serialized_start=1425
  _COMMUNICATIONUSAGE._serialized_end=1638
  _COMMUNICATIONUSAGE_COMMUNICATIONBACKENDTYPE._serialized_start=1536
  _COMMUNICATIONUSAGE_COMMUNICATIONBACKENDTYPE._serialized_end=1638
  _NODEUSAGE._serialized_start=1641
  _NODEUSAGE._serialized_end=1839
  _PROCESSUSAGE._serialized_start=1842
  _PROCESSUSAGE._serialized_end=2459
  _PROCESSUSAGE_RUNTIME._serialized_start=2343
  _PROCESSUSAGE_RUNTIME._serialized_end=2387
  _PROCESSUSAGE_MLFRAMEWORK._serialized_start=2389
  _PROCESSUSAGE_MLFRAMEWORK._serialized_end=2459
  _DEVICEUSAGE._serialized_start=2462
  _DEVICEUSAGE._serialized_end=2930
  _OPSTATS._serialized_start=2933
  _OPSTATS._serialized_end=3423
  _KERNELSTATS._serialized_start=3426
  _KERNELSTATS._serialized_end=3616
  _PROFILERERROR._serialized_start=3618
  _PROFILERERROR._serialized_end=3671
# @@protoc_insertion_point(module_scope)



import unittest
import logging
import sys
import time
from unittest.mock import patch, Mock

import graphsignal
from graphsignal.step_counter import reset_all_step_stats, get_step_stats, update_step_stats

logger = logging.getLogger('graphsignal')


class StepCounterTest(unittest.TestCase):
    def setUp(self):
        graphsignal.configure(
            api_key='k1',
            workload_name='w1',
            debug_mode=True)

    def tearDown(self):
        graphsignal.shutdown()

    def test_update(self):
        reset_all_step_stats()
        
        ss = get_step_stats(1)
        self.assertEqual(ss.step_count, 0)
        self.assertEqual(ss.total_time_us, 0)

        ss = update_step_stats(1, 10, effective_batch_size=128)
        ss = update_step_stats(1, 20, effective_batch_size=128)
        self.assertEqual(ss.step_count, 2)
        self.assertEqual(ss.total_time_us, 30)
        self.assertEqual(ss.sample_count, 2 * 128)

        ss = get_step_stats(2)
        self.assertEqual(ss.step_count, 0)
        self.assertEqual(ss.total_time_us, 0)
        self.assertEqual(ss.sample_count, 0)



from typing import Optional
import logging
import os
import sys
import time
import tempfile
import gzip
import shutil
import glob
import torch
from torch.autograd import DeviceType
import torch.distributed

import graphsignal
from graphsignal.proto_utils import parse_semver
from graphsignal.proto import profiles_pb2
from graphsignal.profiling_step import ProfilingStep
from graphsignal.profilers.framework_profiler import FrameworkProfiler

logger = logging.getLogger('graphsignal')


class PyTorchProfiler(FrameworkProfiler):
    def __init__(self):
        self._torch_prof = None
        self._log_dir = None
        self._ml_framework = None
        self._ml_framework_version = None
        self._global_rank = None
        self._world_size = None
        self._comm_backend = None

    def start(self, profile):
        logger.debug('Activating PyTorch profiler')

        if not self._torch_prof:
            # Initialization
            logger.debug('Warming up PyTorch profiler before first use')

            def _schedule_func(step):
                return torch.profiler.ProfilerAction.RECORD

            self._torch_prof = torch.profiler.profile(
                schedule=_schedule_func,
                record_shapes=False,
                profile_memory=True,
                with_stack=False,
                with_flops=True)
            self._torch_prof.start()
            self._torch_prof.stop()
            logger.debug('Finished warming up')

            self._ml_framework = profiles_pb2.ProcessUsage.MLFramework.Value('PYTORCH')
            self._ml_framework_version = profiles_pb2.SemVer()
            parse_semver(self._ml_framework_version, torch.__version__)

            if torch.distributed.is_available():
                if torch.distributed.is_initialized():
                    self._global_rank = torch.distributed.get_rank()
                    self._world_size = torch.distributed.get_world_size()
                    self._comm_backend = torch.distributed.get_backend()

        # Process info
        profile.process_usage.ml_framework = self._ml_framework
        profile.process_usage.ml_framework_version.CopyFrom(
            self._ml_framework_version)
        if self._global_rank is not None and self._global_rank >= 0:
            if graphsignal._agent.global_rank == -1:
                profile.process_usage.global_rank = self._global_rank

        # Step stats
        if self._world_size is not None and self._world_size > 0:
            profile.step_stats.world_size = self._world_size
            graphsignal.log_parameter('world_size', self._world_size)

        # Communication info
        if self._comm_backend:
            if self._comm_backend == 'nccl':
                profile.comm_usage.backend_type = profiles_pb2.CommunicationUsage.CommunicationBackendType.NCCL
            if self._comm_backend == 'gloo':
                profile.comm_usage.backend_type = profiles_pb2.CommunicationUsage.CommunicationBackendType.GLOO
            if self._comm_backend == 'mpi':
                profile.comm_usage.backend_type = profiles_pb2.CommunicationUsage.CommunicationBackendType.MPI

        self._torch_prof.start()

    def stop(self, profile):
        logger.debug('Deactivating PyTorch profiler')

        self._torch_prof.stop()

        self._convert_operations(profile)

        # Chrome trace
        try:
            self._create_log_dir()
            trace_path = os.path.join(self._log_dir, 'trace.json')
            self._torch_prof.export_chrome_trace(trace_path)
            with open(trace_path) as f:
                trace_json = f.read()
                profile.trace_data = gzip.compress(trace_json.encode())
        except Exception as e:
            logger.error('Error exporting Chrome trace', exc_info=True)
        finally:
            self._remove_log_dir()

    def _create_log_dir(self):
        self._log_dir = tempfile.mkdtemp(prefix='graphsignal-')
        logger.debug('Created temporary log directory %s', self._log_dir)

    def _remove_log_dir(self):
        shutil.rmtree(self._log_dir)
        logger.debug('Removed temporary log directory %s', self._log_dir)

    def _convert_operations(self, profile):
        # Operation stats
        for event_avg in self._torch_prof.key_averages():
            if event_avg.key and event_avg.key.startswith('ProfilerStep'):
                continue
            op_stats = profile.op_stats.add()
            op_stats.op_name = event_avg.key
            op_stats.count = _uint(event_avg.count)
            op_stats.total_host_time_us = _uint(event_avg.cpu_time_total)
            op_stats.total_device_time_us = _uint(event_avg.cuda_time_total)
            op_stats.self_host_time_us = _uint(event_avg.self_cpu_time_total)
            op_stats.self_device_time_us = _uint(event_avg.self_cuda_time_total)
            op_stats.total_host_memory = _uint(event_avg.cpu_memory_usage)
            op_stats.total_device_memory = _uint(event_avg.cuda_memory_usage)
            op_stats.self_host_memory = _uint(event_avg.self_cpu_memory_usage)
            op_stats.self_device_memory = _uint(event_avg.self_cuda_memory_usage)
            op_stats.flops = _uint(event_avg.flops)
            if event_avg.device_type in (DeviceType.CUDA, DeviceType.HIP)  or op_stats.self_device_time_us > 0:
                op_stats.device_type = profiles_pb2.DeviceType.GPU
            else:
                op_stats.device_type = profiles_pb2.DeviceType.CPU

        # Kernel stats
        kernel_index = {}
        for event in self._torch_prof.events():
            for kernel in event.kernels:
                key = (event.key, kernel.name, kernel.device)
                if key in kernel_index:
                    kernel_stats = kernel_index[key]
                    kernel_stats.count += 1
                    kernel_stats.duration_ns += _uint(kernel.duration * 1000)
                else:
                    kernel_stats = kernel_index[key] = profile.kernel_stats.add()
                    kernel_stats.device_type = profiles_pb2.DeviceType.GPU
                    kernel_stats.device_id = str(kernel.device)
                    kernel_stats.op_name = event.name
                    kernel_stats.kernel_name = kernel.name
                    kernel_stats.count = 1
                    kernel_stats.duration_ns = _uint(kernel.duration * 1000)

        for kernel_stats in kernel_index.values():
            profile.kernel_stats.append(kernel_stats)

        logger.debug(
            'Converted %d PyTorch operation statistics', len(profile.op_stats))

def _uint(val):
    return max(int(val), 0)


_profiler = PyTorchProfiler()

def profile_step(
        phase_name: Optional[str] = None,
        effective_batch_size: Optional[int] = None,
        ensure_profile: Optional[bool] = False) -> ProfilingStep:
    graphsignal._check_configured()

    return ProfilingStep(
        phase_name=phase_name,
        effective_batch_size=effective_batch_size,
        ensure_profile=ensure_profile,
        framework_profiler=_profiler)



from typing import Union
import logging
import time
from threading import Lock
import traceback

import graphsignal
from graphsignal.proto import profiles_pb2
from graphsignal.profile_scheduler import select_scheduler
from graphsignal.step_counter import get_step_stats, update_step_stats

logger = logging.getLogger('graphsignal')


class ProfilingStep:
    __slots__ = [
        '_scheduler',
        '_framework_profiler',
        '_is_scheduled',
        '_is_profiling',
        '_profile',
        '_stop_lock',
        '_phase_name',
        '_effective_batch_size',
        '_start_us',
        '_metrics'
    ]

    def __init__(self, 
            phase_name=None, 
            effective_batch_size=None, ensure_profile=False, 
            framework_profiler=None):
        if phase_name is not None:
            if not isinstance(phase_name, str):
                raise ValueError('Invalid phase_name')
            phase_name = phase_name[:50]
        self._phase_name = phase_name

        if effective_batch_size is not None and not isinstance(effective_batch_size, int):
                raise ValueError('Invalid effective_batch_size')
        self._effective_batch_size = effective_batch_size

        self._scheduler = select_scheduler(phase_name)
        self._framework_profiler = framework_profiler
        self._is_scheduled = False
        self._is_profiling = False
        self._profile = None
        self._stop_lock = Lock()
        self._metrics = None

        if self._scheduler.lock(ensure=ensure_profile):
            self._is_scheduled = True
            self._profile = profiles_pb2.MLProfile()
            self._profile.workload_name = graphsignal._agent.workload_name
            self._profile.worker_id = graphsignal._agent.worker_id
            self._profile.run_id = graphsignal._agent.run_id
            if phase_name:
                self._profile.phase_name = phase_name
            self._profile.node_usage.node_rank = graphsignal._agent.node_rank 
            self._profile.process_usage.global_rank = graphsignal._agent.global_rank 
            self._profile.process_usage.local_rank = graphsignal._agent.local_rank 
            self._profile.process_usage.start_ms = graphsignal._agent.start_ms

            if not graphsignal._agent.disable_fwk_profiler and self._framework_profiler:
                try:
                    self._framework_profiler.start(self._profile)
                    self._is_profiling = True
                except Exception as exc:
                    self._scheduler.unlock()
                    self._is_profiling = False
                    _add_exception(self._profile, exc)
                    logger.error('Error starting profiler', exc_info=True)

            self._profile.start_us = _timestamp_us()

        self._start_us = _timestamp_us()

    def __enter__(self):
        return self

    def __exit__(self, *exc):
        self.stop()

    def stop(self) -> None:
        stop_us = _timestamp_us()
        with self._stop_lock:
            step_stats = update_step_stats(
                self._phase_name,
                stop_us - self._start_us,
                effective_batch_size=self._effective_batch_size)

            if self._is_scheduled:
                self._profile.end_us = stop_us

                self._profile.step_stats.step_count = step_stats.step_count
                self._profile.step_stats.sample_count = step_stats.sample_count
                self._profile.step_stats.total_time_us = step_stats.total_time_us

                if graphsignal._agent.tags is not None:
                    for value in graphsignal._agent.tags.keys():
                        tag = self._profile.tags.add()
                        tag.value = value

                if graphsignal._agent.params is not None:
                    for name, value in graphsignal._agent.params.items():
                        param = self._profile.params.add()
                        param.name = name
                        param.value = value

                if graphsignal._agent.metrics is not None:
                    for name, value in graphsignal._agent.metrics.items():
                        metric = self._profile.metrics.add()
                        metric.name = name
                        metric.value = value

                if self._is_profiling:
                    try:
                        self._framework_profiler.stop(self._profile)
                    except Exception as exc:
                        logger.error('Error stopping profiler', exc_info=True)
                        _add_exception(self._profile, exc)

                try:
                    graphsignal._agent.process_reader.read(self._profile)
                    graphsignal._agent.nvml_reader.read(self._profile)
                except Exception as exc:
                    logger.error('Error reading usage information', exc_info=True)
                    _add_exception(self._profile, exc)

                _upload_profile(self._profile)
                self._is_scheduled = False
                self._is_profiling = False
                self._profile = None
                self._scheduler.unlock()

    def set_effective_batch_size(self, effective_batch_size: int) -> None:
        if not isinstance(effective_batch_size, int):
            raise ValueError('Invalid effective_batch_size')
        self._effective_batch_size = effective_batch_size


def _add_exception(profile, exc):
    profiler_error = profile.profiler_errors.add()
    profiler_error.message = str(exc)
    if exc.__traceback__:
        frames = traceback.format_tb(exc.__traceback__)
        if len(frames) > 0:
            profiler_error.stack_trace = ''.join(frames)


def _upload_profile(profile):
    graphsignal._agent.uploader.upload_profile(profile)
    graphsignal._agent.uploader.flush_in_thread()


def _timestamp_us():
    return int(time.time() * 1e6)



import unittest
import logging
import sys
import time
from unittest.mock import patch, Mock
from google.protobuf.json_format import MessageToJson
import pprint

import graphsignal
from graphsignal.profilers.generic import profile_step
from graphsignal.proto import profiles_pb2
from graphsignal.uploader import Uploader

logger = logging.getLogger('graphsignal')


class GenericProfilerTest(unittest.TestCase):
    def setUp(self):
        if len(logger.handlers) == 0:
            logger.addHandler(logging.StreamHandler(sys.stdout))
        graphsignal.configure(
            api_key='k1',
            workload_name='w1',
            debug_mode=True)

    def tearDown(self):
        graphsignal.shutdown()

    @patch.object(Uploader, 'upload_profile')
    def test_profile_step(self, mocked_upload_profile):
        with profile_step(phase_name='training', ensure_profile=True):
            time.sleep(0.1)

        profile = mocked_upload_profile.call_args[0][0]

        #pp = pprint.PrettyPrinter()
        #pp.pprint(MessageToJson(profile))

        self.assertEqual(profile.phase_name, 'training')
        self.assertTrue(profile.step_stats.step_count > 0)
        self.assertTrue(profile.step_stats.total_time_us > 0)



import logging
from transformers import TrainerCallback

import graphsignal
from graphsignal.proto import profiles_pb2
from graphsignal.profiling_step import ProfilingStep

logger = logging.getLogger('graphsignal')

PHASE_TRAINING = 'training'

EXCLUDE_ARGS = {
    'logging_dir',
    'local_rank'
}

class GraphsignalCallback(TrainerCallback):
    def __init__(self, framework_profiler):
        super().__init__()
        self._profiler = framework_profiler
        self._step = None
        self._local_rank = -1

    def on_train_begin(self, args, state, control, **kwarg):
        self._configure_profiler(args)

    def on_train_end(self, args, state, control, **kwarg):
        self._stop_profiler(args, state)

    def on_step_begin(self, args, state, control, **kwarg):
        self._stop_profiler(args, state)
        self._start_profiler(PHASE_TRAINING, args, state)

    def on_step_end(self, args, state, control, **kwarg):
        pass

    def _start_profiler(self, phase_name, args, state):
        if not self._step:
            self._step = ProfilingStep(
                phase_name=phase_name,
                effective_batch_size=self._get_effective_batch_size(args),
                framework_profiler=self._profiler)

    def _stop_profiler(self, args, state):
        if self._step:
            if self._step._is_scheduled:
                self._update_profile(args, state)
            self._step.stop()
            self._step = None

    def _get_effective_batch_size(self, args):
        gradient_accumulation_steps = args.gradient_accumulation_steps if args.gradient_accumulation_steps > 0 else 1
        return args.train_batch_size * args.gradient_accumulation_steps

    def _update_profile(self, args, state):
        profile = self._step._profile

        if args.local_rank == -1 or args.local_rank == 0:
            profile.step_stats.flop_count = _uint(state.total_flos)
        profile.step_stats.batch_size = args.train_batch_size
        profile.step_stats.device_batch_size = args.per_device_train_batch_size

        if self._local_rank >= 0 and graphsignal._agent.local_rank == -1:
            profile.process_usage.local_rank = self._local_rank

    def _configure_profiler(self, args):
        if args.local_rank >= 0:
            self._local_rank = args.local_rank

        if args.world_size > 0:
            graphsignal.log_parameter('world_size', args.world_size)

        for name, value in vars(args).items():
            if not name.startswith('_') and name not in EXCLUDE_ARGS and isinstance(value, (str, int, float, bool)):
                graphsignal.log_parameter(name, value)


def _uint(val):
    return max(int(val), 0)


class GraphsignalPTCallback(GraphsignalCallback):
    def __init__(self):
        from graphsignal.profilers.pytorch import PyTorchProfiler
        super().__init__(PyTorchProfiler())


class GraphsignalTFCallback(GraphsignalCallback):
    def __init__(self):
        from graphsignal.profilers.tensorflow import TensorflowProfiler
        super().__init__(TensorflowProfiler())



from typing import Any, Union, Optional
import time
import sys
import os
import logging
import threading
import uuid
import hashlib
import atexit

from graphsignal import version
from graphsignal.agent import Agent
from graphsignal.uploader import Uploader
from graphsignal.usage.process_reader import ProcessReader
from graphsignal.usage.nvml_reader import NvmlReader
from graphsignal.proto import profiles_pb2

logger = logging.getLogger('graphsignal')

_agent = None


def _check_configured():
    global _agent
    if not _agent:
        raise ValueError(
            'Graphsignal profiler not configured, call graphsignal.configure() first')


def _check_and_set_arg(
        name, value, is_str=False, is_int=False, is_bool=False, required=True):
    env_name = 'GRAPHSIGNAL_{0}'.format(name.upper())

    if not value and env_name in os.environ:
        value = os.environ[env_name]
        if value:
            if is_int:
                try:
                    value = int(value)
                except:
                    raise ValueError('Invalid format, expected integer: {0}'.format(name))
            elif is_bool:
                value = bool(value)

    if not value:
        if required:
            raise ValueError('Missing argument: {0}'.format(name))
    else:
        if is_str:
            if not isinstance(value, str):
                raise ValueError('Invalid format, expected string: {0}'.format(name))
        elif is_int:
            if not isinstance(value, int):
                raise ValueError('Invalid format, expected integer: {0}'.format(name))
        elif is_bool:
            if not isinstance(value, bool):
                raise ValueError('Invalid format, expected boolean: {0}'.format(name))

    return value


def configure(
        api_key: str = None,
        workload_name: str = None,
        run_id: Optional[str] = None,
        global_rank: Optional[int] = None,
        node_rank: Optional[int] = None,
        local_rank: Optional[int] = None,
        debug_mode: Optional[bool] = False,
        disable_fwk_profiler: Optional[bool] = False) -> None:
    global _agent

    if _agent:
        logger.warning('Graphsignal profiler already configured')
        return

    debug_mode = _check_and_set_arg('debug_mode', debug_mode, is_bool=True, required=False)
    if debug_mode:
        logger.setLevel(logging.DEBUG)
    else:
        logger.setLevel(logging.WARNING)
    api_key = _check_and_set_arg('api_key', api_key, is_str=True, required=True)
    workload_name = _check_and_set_arg('workload_name', workload_name,  is_str=True, required=True)
    run_id = _check_and_set_arg('run_id', run_id, is_str=True, required=False)
    global_rank = _check_and_set_arg('global_rank', global_rank, is_int=True, required=False)
    node_rank = _check_and_set_arg('node_rank', node_rank, is_int=True, required=False)
    local_rank = _check_and_set_arg('local_rank', local_rank, is_int=True, required=False)
    disable_fwk_profiler = _check_and_set_arg('disable_fwk_profiler', disable_fwk_profiler, is_bool=True, required=False)

    if not run_id:
        run_id = _uuid_sha1()
        # set run ID to be picked up by worker processes
        os.environ['GRAPHSIGNAL_RUN_ID'] = run_id
        logger.debug('run_id (or GRAPHSIGNAL_RUN_ID) not available, generated: %s', run_id)
    else:
        logger.debug('run_id (or GRAPHSIGNAL_RUN_ID) available: %s', run_id)

    _agent = Agent()
    _agent.worker_id = _uuid_sha1(size=12)
    _agent.start_ms = int(time.time() * 1e3)
    _agent.api_key = api_key
    _agent.workload_name = workload_name[:50]
    _agent.run_id = _sha1(run_id, size=12)
    _agent.node_rank = node_rank if node_rank is not None else -1
    _agent.local_rank = local_rank if local_rank is not None else -1
    _agent.global_rank = global_rank if global_rank is not None else -1
    _agent.debug_mode = debug_mode
    _agent.uploader = Uploader()
    _agent.uploader.configure()
    _agent.process_reader = ProcessReader()
    _agent.process_reader.setup()
    _agent.nvml_reader = NvmlReader()
    _agent.nvml_reader.setup()

    atexit.register(shutdown)

    logger.debug('Graphsignal profiler configured')


def add_tag(tag: str) -> None:
    _check_configured()

    if tag is None or not isinstance(tag, str):
        raise ValueError('Missing or invalid argument: tag')

    global _agent
    if _agent.tags is None:
        _agent.tags = {}
    _agent.tags[tag[:50]] = True


def log_parameter(name: str, value: Any) -> None:
    _check_configured()

    if name is None or not isinstance(name, str):
        raise ValueError('Missing or invalid argument: name')

    if value is None:
        raise ValueError('Missing argument: value')

    global _agent
    if _agent.params is None:
        _agent.params = {}
    _agent.params[name[:250]] = str(value)[:1000]


def log_metric(name: str, value: Union[int, float]) -> None:
    _check_configured()

    if name is None or not isinstance(name, str):
        raise ValueError('Missing or invalid argument: name')

    if value is None or not isinstance(value, (int, float)):
        raise ValueError('Missing argument: value')

    if _agent.metrics is None:
        _agent.metrics = {}
    _agent.metrics[name[:250]] = value


def shutdown() -> None:
    _check_configured()

    global _agent
    atexit.unregister(shutdown)
    _agent.uploader.flush()
    _agent.process_reader.shutdown()
    _agent.nvml_reader.shutdown()
    _agent = None

    logger.debug('Graphsignal profiler shutdown')


def generate_uuid() -> None:
    return _uuid_sha1()    


def _sha1(text, size=-1):
    sha1_hash = hashlib.sha1()
    sha1_hash.update(text.encode('utf-8'))
    return sha1_hash.hexdigest()[0:size]


def _uuid_sha1(size=-1):
    return _sha1(str(uuid.uuid4()), size)



from typing import Optional
import logging

from tensorflow.keras.callbacks import Callback

import graphsignal
from graphsignal.proto import profiles_pb2
from graphsignal.profilers.tensorflow import TensorflowProfiler
from graphsignal.profiling_step import ProfilingStep

logger = logging.getLogger('graphsignal')

PHASE_TRAINING = 'training'
PHASE_TEST = 'test'
PHASE_PREDICTION = 'prediction'


class GraphsignalCallback(Callback):
    def __init__(self, batch_size: Optional[int] = None):
        super().__init__()
        self._profiler = TensorflowProfiler()
        self._step = None
        self._batch_size = batch_size

    def on_train_begin(self, logs=None):
        self._configure_profiler()

    def on_train_end(self, logs=None):
        self._stop_profiler()

    def on_test_begin(self, logs=None):
        self._configure_profiler()

    def on_test_end(self, logs=None):
        self._stop_profiler()

    def on_predict_begin(self, logs=None):
        self._configure_profiler()

    def on_predict_end(self, logs=None):
        self._stop_profiler()

    def on_train_batch_begin(self, batch, logs=None):
        self._stop_profiler()
        self._start_profiler(PHASE_TRAINING)

    def on_train_batch_end(self, batch, logs=None):
        self._log_metrics(logs)

    def on_test_batch_begin(self, batch, logs=None):
        self._stop_profiler()
        self._start_profiler(PHASE_TEST)

    def on_test_batch_end(self, batch, logs=None):
        self._log_metrics(logs)

    def on_predict_batch_begin(self, batch, logs=None):
        self._stop_profiler()
        self._start_profiler(PHASE_PREDICTION)

    def on_predict_batch_end(self, batch, logs=None):
        self._log_metrics(logs)

    def _configure_profiler(self):
        if self._batch_size:
            graphsignal.log_parameter('batch_size', self._batch_size)

    def _start_profiler(self, phase_name):
        if not self._step:
            self._step = ProfilingStep(
                phase_name=phase_name,
                effective_batch_size=self._batch_size,
                framework_profiler=self._profiler)

    def _stop_profiler(self):
        if self._step:
            if self._step._is_scheduled:
                self._update_profile()
            self._step.stop()
            self._step = None

    def _update_profile(self):
        profile = self._step._profile

        if self._batch_size:
            profile.step_stats.batch_size = self._batch_size

    def _log_metrics(self, logs):
        if logs:
            for key, value in logs.items():
                if isinstance(key, str) and isinstance(value, (int, float)):
                    graphsignal.log_metric(key, value)



from typing import Optional
import logging
import os
import sys
import time
import tempfile
import gzip
import shutil
import glob
import json
import tensorflow as tf
from tensorflow.python.client import device_lib
from google.protobuf.json_format import Parse

from graphsignal.profilers.tensorflow_proto import overview_page_pb2
from graphsignal.profilers.tensorflow_proto import input_pipeline_pb2
from graphsignal.profilers.tensorflow_proto import tf_stats_pb2
from graphsignal.profilers.tensorflow_proto import kernel_stats_pb2
from graphsignal.profilers.tensorflow_proto import memory_profile_pb2

import graphsignal
from graphsignal.proto_utils import parse_semver, compare_semver
from graphsignal.proto import profiles_pb2
from graphsignal.profiling_step import ProfilingStep
from graphsignal.profilers.framework_profiler import FrameworkProfiler

logger = logging.getLogger('graphsignal')


class TensorflowProfiler(FrameworkProfiler):
    def __init__(self):
        self._is_initialized = False
        self._log_dir = None
        self._ml_framework = None
        self._ml_framework_version = None
        self._global_rank = None
        self._world_size = None

    def start(self, profile):
        logger.debug('Activating TensorFlow profiler')

        # Initialization
        if not self._is_initialized:
            logger.debug('Warming up TensorFlow profiler before first use')
            self._is_initialized = True
            tf.profiler.experimental.start('')
            tf.profiler.experimental.stop(save=False)
            logger.debug('Finished warming up')

            self._ml_framework = profiles_pb2.ProcessUsage.MLFramework.TENSORFLOW
            self._ml_framework_version = profiles_pb2.SemVer()
            parse_semver(self._ml_framework_version, tf.__version__)
            if compare_semver(self._ml_framework_version, (2, 2, 0)) == -1:
                raise Exception(
                    'TensorFlow profiling is not supported for versions <=2.2')

            if 'TF_CONFIG' in os.environ:
                try:
                    tf_config = json.loads(os.environ['TF_CONFIG'])
                    self._world_size = 0
                    if 'chief' in tf_config['cluster']:
                        self._world_size += len(tf_config['cluster']['chief'])
                    if 'worker' in tf_config['cluster']:
                        self._world_size += len(tf_config['cluster']['worker'])
                    self._global_rank = tf_config['task']['index']
                except:
                    logger.warning('Error parsing TF_CONFIG', exc_info=True)

        # Process info
        profile.process_usage.ml_framework = self._ml_framework
        profile.process_usage.ml_framework_version.CopyFrom(
            self._ml_framework_version)
        if self._global_rank is not None and self._global_rank >= 0:
            if graphsignal._agent.global_rank == -1:
                profile.process_usage.global_rank = self._global_rank

        # Step stats
        if self._world_size is not None and self._world_size > 0:
            profile.step_stats.world_size = self._world_size
            graphsignal.log_parameter('world_size', self._world_size)

        try:
            self._create_log_dir()

            options = tf.profiler.experimental.ProfilerOptions(
                host_tracer_level=2,
                python_tracer_level=1,
                device_tracer_level=1)
            tf.profiler.experimental.start(self._log_dir, options=options)
        except Exception as e:
            self._remove_log_dir()
            raise e

    def stop(self, profile):
        logger.debug('Deactivating TensorFlow profiler')

        try:
            tf.profiler.experimental.stop()

            self._convert_operations(profile)

            trace_json_gz = self._find_and_read(
                'plugins/profile/*/*trace.json.gz',
                decompress=False)
            profile.trace_data = trace_json_gz
        except Exception as e:
            raise e
        finally:
            self._remove_log_dir()

    def _create_log_dir(self):
        self._log_dir = tempfile.mkdtemp(prefix='graphsignal-')
        logger.debug('Created temporary log directory %s', self._log_dir)

    def _remove_log_dir(self):
        shutil.rmtree(self._log_dir)
        logger.debug('Removed temporary log directory %s', self._log_dir)

    def _convert_operations(self, profile):
        # Operation stats
        tf_stats_data = self._find_and_read(
            'plugins/profile/*/*tensorflow_stats.pb')
        if tf_stats_data:
            tf_stats_db = tf_stats_pb2.TfStatsDatabase()
            tf_stats_db.ParseFromString(tf_stats_data)
            for tf_stats_record in tf_stats_db.without_idle.tf_stats_record:
                op_stats = profile.op_stats.add()
                if tf_stats_record.host_or_device == 'Host':
                    op_stats.device_type = profiles_pb2.DeviceType.CPU
                    op_stats.total_host_time_us = _uint(tf_stats_record.total_time_in_us)
                    op_stats.self_host_time_us = _uint(tf_stats_record.total_self_time_in_us)
                    op_stats.self_host_memory_rate = _uint(tf_stats_record.measured_memory_bw)
                else:
                    op_stats.device_type = profiles_pb2.DeviceType.GPU
                    op_stats.total_device_time_us = _uint(tf_stats_record.total_time_in_us)
                    op_stats.self_device_time_us = _uint(tf_stats_record.total_self_time_in_us)
                    op_stats.self_device_memory_rate = _uint(tf_stats_record.measured_memory_bw)
                    op_stats.tensorcore_utilization = tf_stats_record.gpu_tensorcore_utilization
                op_stats.op_type = tf_stats_record.op_type
                op_stats.op_name = tf_stats_record.op_name
                op_stats.count = _uint(tf_stats_record.occurrences)
                op_stats.flops_per_sec = _uint(tf_stats_record.measured_flop_rate)
        else:
            logger.debug('No operation data found in TensorFlow log directory')

        # Kernel stats
        kernel_stats_data = self._find_and_read(
            'plugins/profile/*/*kernel_stats.pb')
        if kernel_stats_data:
            kernel_stats_db = kernel_stats_pb2.KernelStatsDb()
            kernel_stats_db.ParseFromString(kernel_stats_data)
            for kernel_report in kernel_stats_db.reports:
                kernel_stats = profile.kernel_stats.add()
                kernel_stats.device_type = profiles_pb2.DeviceType.GPU
                kernel_stats.op_name = kernel_report.op_name
                kernel_stats.kernel_name = kernel_report.name
                kernel_stats.count = _uint(kernel_report.occurrences)
                kernel_stats.duration_ns = _uint(kernel_report.total_duration_ns)
                kernel_stats.is_using_tensorcore = kernel_report.is_kernel_using_tensor_core
        else:
            logger.debug('No kernel data found in TensorFlow log directory')

    def _find_and_read(self, file_pattern, decompress=True):
        file_paths = glob.glob(os.path.join(self._log_dir, file_pattern))
        if len(file_paths) == 0:
            raise Exception(
                'Files are not found at {}'.format(
                    os.path.join(
                        self._log_dir,
                        file_pattern)))

        if decompress and file_paths[-1].endswith('.gz'):
            last_file = gzip.open(file_paths[-1], "rb")
        else:
            last_file = open(file_paths[-1], "rb")
        data = last_file.read()
        last_file.close()

        return data


def _uint(val):
    return max(int(val), 0)


_profiler = TensorflowProfiler()

def profile_step(
        phase_name: Optional[str] = None,
        effective_batch_size: Optional[int] = None,
        ensure_profile: Optional[bool] = False) -> ProfilingStep:
    graphsignal._check_configured()

    return ProfilingStep(
        phase_name=phase_name,
        effective_batch_size=effective_batch_size,
        ensure_profile=ensure_profile,
        framework_profiler=_profiler)
